#
# Copyright (C) 2019 Sylvain Marsat.
#
#


"""
    Cython wrapping of numerical constants.
"""

import numpy as np
cimport numpy as np


from lisabeta.pyconstants cimport *

SUCCESS = _SUCCESS
FAILURE = _FAILURE

PI = _PI
PI_2 = _PI_2
PI_3 = _PI_3
PI_4 = _PI_4
SQRTPI = _SQRTPI
SQRTTWOPI = _SQRTTWOPI
INVSQRTPI = _INVSQRTPI
INVSQRTTWOPI = _INVSQRTTWOPI
GAMMA = _GAMMA
SQRT2 = _SQRT2
SQRT3 = _SQRT3
SQRT6 = _SQRT6
INVSQRT2 = _INVSQRT2
INVSQRT3 = _INVSQRT3
INVSQRT6 = _INVSQRT6

C_SI = _C_SI
G_SI = _G_SI
MSUN_SI = _MSUN_SI
MTSUN_SI = _MTSUN_SI
MRSUN_SI = _MRSUN_SI
PC_SI = _PC_SI
AU_SI = _AU_SI
YRSID_SI = _YRSID_SI

EarthOrbitOmega_SI = _EarthOrbitOmega_SI

EPOCH_J2000_0_TAI_UTC = _EPOCH_J2000_0_TAI_UTC
EPOCH_J2000_0_GPS = _EPOCH_J2000_0_GPS
EPOCH_LISA_0_GPS = _EPOCH_LISA_0_GPS
